/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/*
** VID_GetModeInfo
*/

#define VID_NUM_MODES ( sizeof( vid_modes ) / sizeof( vid_modes[0] ) )

typedef struct vidmode_s {
	const char *description;
	int         width, height;
	int         mode;
} vidmode_t;

static const vidmode_t vid_modes[] = {
#ifdef _WIN32_WCE
	{ "Mode 0: 240x320",	240,  300,	 0 },
#else
	{ "Mode 0: 320x240",	320,  240,	 0 },
#endif
	{ "Mode 1: 400x300",	400,  300,	 1 },
	{ "Mode 2: 512x384",	512,  384,	 2 },
	{ "Mode 3: 640x480",	640,  480,	 3 },
	{ "Mode 4: 800x600",	800,  600,	 4 },
	{ "Mode 5: 960x720",	960,  720,	 5 },
	{ "Mode 6: 1024x768",	1024, 768,	 6 },
	{ "Mode 7: 1152x864",	1152, 864,	 7 },
	{ "Mode 8: 1280x960",	1280, 960,	 8 },
	{ "Mode 9: 1600x1200",	1600, 1200,	 9 },
	{ "Mode 10: 2048x1536",	2048, 1536,	10 },
	{ "Mode 11: 1280x1024",	1280, 1024,	11 },
	{ "Mode 12: 1280x800",	1280, 800,	12 },
};

static qboolean VID_GetModeInfo( int *width, int *height, int mode ) {
	if ( mode < 0 || mode >= VID_NUM_MODES )
		return qfalse;

	*width  = vid_modes[mode].width;
	*height = vid_modes[mode].height;

	return qtrue;
}

